/****************************************************************************
 *
 *   File: MLGraph.cxx:
 *   
 *   Author: Mark Stephenson
 *   Date: Mon Dec  1 13:59:36 2003
 *
 *   Function: A dense matrix graph representation.  The ORC graph is hard
 *      to query and modify, so I'm making my own.
 *
 ***************************************************************************/

#include <math.h>
#include <stdarg.h>
#include <set>

#include "defs.h"
#include "config.h"
#include "errors.h"
#include "mempool.h"
#include "cg_flags.h"
#include "cgir.h"
#include "tracing.h"
#include "config_asm.h"
#include "tn_map.h"
#include "cio_rwtran.h"
#include "cgprep.h"
#include "op.h"
#include "op_list.h"
#include "bb.h"
#include "cgtarget.h"
#include "cg_swp_target.h"
#include "wn.h"
#include "wn_core.h"
#include "wn_util.h"
#include "whirl2ops.h"

#include "cg.h"
#include "register.h"	/* needed for "gra.h" */
#include "tn_set.h"	/* needed for "gra.h" */
#include "gra.h"
#include "gra_live.h"
#include "bb_set.h"
#include "gtn_universe.h"
#include "gtn_set.h"
#include "pf_cg.h"
#include "note.h"
#include "cgexp.h"
#include "cio.h"
#include "cio_rwtran.h"
#include "cg_cflow.h"
#include "resource.h"
#include "cg_db_op.h"
#include "dominate.h"
#include "ti_res_count.h"
#include "cg_loop_scc.h"
#include "cg_loop_recur.h"
#include "cg_loop_mii.h"
#include "freq.h"
#include "cg_region.h"
#include "cg_sched_est.h"
#include "cg_loop.h"
#include "cg_swp.h"
#include "cg_swp_options.h"
#include "findloops.h"
#include "dominate.h"
#include "ebo.h"
#include "hb.h"
#include "gra_live.h"
#include "ipfec_options.h"
#include "if_conv.h"
#include "region_bb_util.h"
#include "MLGraph.h"
#include "cg_dep_graph.h"
#include "dag.h"
#include <fstream.h>

#include "targ_issue_port.h" // To get PROCESSOR_Version


MLGraph::MLGraph (BB *loophead, bool cyclic) :
  m_cyclic(cyclic),
  m_loophead(loophead),
  m_arith_store(0),
  m_load_arith_store(0),
  m_invariants(0)
//////////////////////////////////////////////////////////////////
//
//  Convert the cg_dep_graph to a regular sparse matrix 
//  representation.
//
//////////////////////////////////////////////////////////////////
{
  OP *op;
  ARC *arc;
  ARC_LIST *arcs;
  UINT32 size;

  CG_DEP_Compute_Graph (loophead,
			NO_ASSIGNED_REG_DEPS, 
			cyclic,
			NO_MEMREAD_ARCS,
			INCLUDE_MEMIN_ARCS,
			INCLUDE_CONTROL_ARCS, 
			NULL);

  // Find out how many ops there are and create and
  // enumerate each of them.
  FOR_ALL_BB_OPs(loophead, op) 
  {
    m_nodes.push_back (op);
  }
 
  size = m_nodes.size ();
  m_edges = new MLEdge[size*size];
  m_height = new UINT64[size];
  m_hallucinate = (ARC*) this;

  for (int i = 0; i < size; i++)
    m_height[i] = 0;

  for (int i = 0; i < size*size; i++)
  {
    m_edges[i].edge = NULL;
    m_edges[i].omega = 0;
  }

  FOR_ALL_BB_OPs(loophead, op) 
  {
    if (_CG_DEP_op_info(op)) 
    {
      for (arcs = OP_succs(op); arcs; arcs = ARC_LIST_rest(arcs)) 
      {
	arc = ARC_LIST_first(arcs);
	if (ARC_kind(arc) != CG_DEP_PREBR)
	{
	  add_edge (arc);
	}
      }
    }
  }
}


MLGraph::~MLGraph ()
//////////////////////////////////////////////////////////////////
//
//  Delete the graph.
//
//////////////////////////////////////////////////////////////////
{
  delete[] m_edges;
  delete[] m_height;

  list<set<UINT32>*>::iterator first = m_sccs.begin ();
  list<set<UINT32>*>::iterator last = m_sccs.end ();
  while (first != last)
  {
    delete (*first);
    first++;
  }

  CG_DEP_Delete_Graph (m_loophead);
}


void
MLGraph::print (INT loopid, char* des)
//////////////////////////////////////////////////////////////////
//
//  Prints the graph -- in dot format -- to a file.
//
//////////////////////////////////////////////////////////////////
{
  char depfname[80];
  sprintf (depfname, "%s_%d.dot", des, loopid);
  FILE *depfile = fopen(depfname, "w");

  fprintf (depfile, "digraph %s_%d {\n", des, loopid);

  for (int i = 0; i < m_nodes.size(); i++)
  {
    OP* op = m_nodes[i];
    fprintf (depfile, "op%d [label=\"%s\"]\n", i, TOP_Name(OP_code(op)));
  }
   
  for (int i = 0; i < m_nodes.size(); i++)
  {
    for (int j = 0; j < m_nodes.size(); j++)
    {
      MLEdge *edge = &(m_edges[get_address(i, j)]);
      if (normal_edge(*edge))
      {
	fprintf (depfile, "op%d -> op%d [label=\"%d\", color=%s]\n", 
		 i, j, edge->omega, edge->omega ? "blue" : "black");
      }
    }
  }

  fprintf (depfile, "}\n");
  fclose (depfile);
}


void
MLGraph::degrees (double& in_degree,
		  double& out_degree)
//////////////////////////////////////////////////////////////////
//
//  Computes the average in-degree and out-degree of the nodes
//  in the graph.  The values are passed by reference.
//
//////////////////////////////////////////////////////////////////
{
  in_degree = 0;
  out_degree = 0;

  for (int i = 0; i < m_nodes.size(); i++)
  {
    for (int j = 0; j < m_nodes.size(); j++)
    {
      if (normal_edge_present(i, j))
	in_degree++;
      
      if (normal_edge_present(j, i))
	out_degree++;
    }
  }
  
  in_degree = in_degree / m_nodes.size();
  out_degree = out_degree / m_nodes.size();  
}


void 
MLGraph::dep_height (UINT32& num_sccs, 
		     UINT32& max_dh, 
		     UINT16& max_memdep,
		     UINT16& max_contdep,
		     double& avg_dh, 
		     UINT32& non_trivial)
//////////////////////////////////////////////////////////////////
//
//  Computes the number of computations (sccs) in the loop, as
//  well as the maximum dependence height, the maximum number
//  of dependent memory operations, the average dependence
//  height, and the number of non-trivial computations.
// 
//  This function should only be called for non-cyclic graphs.
//
//////////////////////////////////////////////////////////////////
{
  list<set<UINT32>*>::iterator first = m_sccs.begin ();
  list<set<UINT32>*>::iterator last = m_sccs.end ();
  double sum = 0.0;
  UINT32 depth;
  UINT16 memdep;
  UINT16 contdep;

  max_dh = 0;
  max_memdep = 0;
  max_contdep = 0;
  non_trivial = 0;
  num_sccs = m_sccs.size ();

  while (first != last)
  {
    max_depth (*first, depth, memdep, contdep);
    sum += (double) depth;
    if (depth > max_dh)
      max_dh = depth;
    if (memdep > max_memdep)
      max_memdep = memdep;
    if (contdep > max_contdep)
      max_contdep = contdep;
    if (depth > 1)
      non_trivial++;
    first++;
  }

  avg_dh = sum / ((double) num_sccs);
}


void
MLGraph::cyclic_sccs (UINT32& num_sccs,
		      UINT32& non_trivial)
//////////////////////////////////////////////////////////////////
//
//  This function computes the number of sccs in the graph
//  and of these, the number of non-trivial sccs.  This function
//  should only be called for cyclic graphs.
//
//////////////////////////////////////////////////////////////////
{
  list<set<UINT32>*>::iterator first = m_sccs.begin ();
  list<set<UINT32>*>::iterator last = m_sccs.end ();

  non_trivial = 0;
  num_sccs = m_sccs.size ();

  while (first != last)
  {
    set<UINT32> *s = *first;
    if (s->size() > 1)
      non_trivial++;
    first++;
  }
}


UINT32
MLGraph::find (OP* op)
//////////////////////////////////////////////////////////////////
//
//  Finds the number that the operation hashes to.
//
//////////////////////////////////////////////////////////////////
{
  for (UINT32 i = 0; i < m_nodes.size(); i++)
    if (m_nodes[i] == op)
      return i;

  return m_nodes.size();
}


void
MLGraph::add_edge (ARC *arc)
//////////////////////////////////////////////////////////////////
//
//  Adds an edge in the matrix representation.
//
//////////////////////////////////////////////////////////////////
{
  BOOL definite;
  UINT8 omega;

  OP *opfrom = ARC_succ (arc);
  OP *opto = ARC_pred (arc);
  
  UINT32 from = find (opfrom);
  UINT32 to = find(opto);

  MLEdge *edge = &(m_edges[get_address(from, to)]);
  edge->edge = arc;

  if (get_mem_dep (opto, opfrom, &definite, &omega))
    edge->omega = omega;

  // Hallucinate the other edge.
  if (from != to)
    m_edges[get_address(to, from)].edge = m_hallucinate;
}


void
MLGraph::depth (UINT32 node, 
		UINT32& dep, 
		UINT16& memdep)
//////////////////////////////////////////////////////////////////
//
//  Computes the depth of a given node.
//
//////////////////////////////////////////////////////////////////
{
  UINT16 mem_height = 0;
  UINT32 height = 0;
  UINT32 h;
  UINT16 m;
  UINT64 memoized;

  memoized = m_height[node];
  if (memoized)
  {
    dep = (UINT32) (memoized & 0xffffffff);
    memdep = ((UINT16) (memoized >> 32)) & 0xffff;
    return;
  }

  for (int i = 0; i < m_nodes.size(); i++)
  {
    if (i != node && normal_edge_present(node, i))
    {
      depth (i, h, m);
      if (h > height)
	height = h;
      if (m > mem_height)
	mem_height = m;
    }
  }

  dep = height + 1;
  memdep = mem_height + (OP_memory(m_nodes[node]) ? 1 : 0);
  m_height[node] = ((UINT64)memdep << 32) | (UINT64)dep;
}


void
MLGraph::max_depth (set<UINT32>* C, 
		    UINT32& height, 
		    UINT16& memdep,
		    UINT16& contdep)
//////////////////////////////////////////////////////////////////
//
//  Finds the depth of a computation.
//
//////////////////////////////////////////////////////////////////
{
  UINT32 h;
  UINT16 md;
  TN    *t;
  UINT32 max_height = 0;
  UINT16 max_memdep = 0;

  set<unsigned> preds_used;
  set<UINT32>::iterator first = C->begin ();
  set<UINT32>::iterator last = C->end ();

  while (first != last)
  {
    depth (*first, h, md);
      
    if (h > max_height)
      max_height = h;
    if (md > max_memdep)
      max_memdep = md;

    t = OP_opnd(m_nodes[*first], OP_PREDICATE_OPND);
    preds_used.insert(TN_number(t));

    first++;
  }

  contdep = preds_used.size ();
  height = max_height;
  memdep = max_memdep;
}


void
MLGraph::find_sccs ()
//////////////////////////////////////////////////////////////////
//
//  Finds the strongly connected components in the graph.
//  For cyclic graphs, only true edges are used; non-cyclic
//  graphs use hallucinated edges-- this allows us to find the
//  number of distinct computations.
//
//////////////////////////////////////////////////////////////////
{
  int *val;
  int *stack;

  val = new int[m_nodes.size()];
  stack = new int[m_nodes.size()];
  for (int i = 0; i < m_nodes.size(); i++)
    val[i] = 0;

  for (int i = 0; i < m_nodes.size(); i++)
  {
    if (val[i] == 0)
    {
      m_id = 0;
      m_p = 0;
      
      find_sccs (val, stack, i);
    }
  }

  delete[] val;
  delete[] stack;
}


int
MLGraph::find_sccs (int *val, int *stack, int k)
//////////////////////////////////////////////////////////////////
//
//  The helper function to find sccs.
//
//////////////////////////////////////////////////////////////////
{
  int m;
  int min;

  min = ++m_id;
  val[k] = m_id;
  stack[m_p++] = k;

  for (int t = 0; t < m_nodes.size(); t++)
  {
    if ((m_cyclic && normal_edge_present(k, t)) ||
	(!m_cyclic && edge_present(k, t)))
    {
      m = (!val[t]) ? find_sccs(val, stack, t) : val[t];
      if (m < min) min = m;
    }
  }

  if (min == val[k])
  {
    set<UINT32> *C = new set<UINT32>;
    do
    {
      C->insert(stack[--m_p]);
      val[stack[m_p]] = m_nodes.size()+1;
    } while (stack[m_p] != k);

    m_sccs.push_back(C);
  }

  return min;
}


UINT32
MLGraph::get_address (UINT32 m, UINT32 n)
//////////////////////////////////////////////////////////////////
//
//  Converts an index into a 2-d array into an index into a
//  1-d array.
//
//////////////////////////////////////////////////////////////////
{
  UINT32 size;

  size = m_nodes.size ();

  assert (m < size);
  assert (n < size);

  return ((size * m) + n);
}


OP*
MLGraph::single_succ (OP *op)
//////////////////////////////////////////////////////////////////
//
//  Returns the OP that follows op if it is the only successor
//  to op.  Otherwise, NULL is returned.
//
//////////////////////////////////////////////////////////////////
{
  ARC *arc;
  ARC_LIST *arcs;
  OP *unique = NULL;

  for (arcs = OP_succs(op); arcs; arcs = ARC_LIST_rest(arcs)) 
  {
    arc = ARC_LIST_first(arcs);
    if (ARC_kind(arc) != CG_DEP_PREBR)
    {
      if (unique) return NULL;
      unique = ARC_succ (arc);
    }
  }

  return unique;
}


void
MLGraph::find_indirect_refs (UINT32& ind_refs)
//////////////////////////////////////////////////////////////////
//
//  Finds the number of indirect references in the graph. 
//  Indirect references are array accesses of the form: a[b[i]].
//
//////////////////////////////////////////////////////////////////
{
  UINT8 opnum;
  UINT8 ixopnd;

  ind_refs = 0;
  for (INT i = 0; i < m_nodes.size(); i++)
  {
    OP *op = m_nodes[i];
    if (OP_memory(op))
    {
      for (INT j = 0; j < m_nodes.size(); j++)
      {
	MLEdge edge = m_edges[get_address(i, j)];
	ARC *arc = edge.edge;
	if (normal_edge(edge) &&
	    (ARC_kind(arc) == CG_DEP_REGIN ||
	     ARC_kind(arc) == CG_DEP_REGANTI))
	{
	  opnum = ARC_opnd(arc);
	  ixopnd = (floating_store(op)) ? 2 : 3;
	  if ((opnum == ixopnd) && 
	      connected_to_load(j))
	  {
	    ind_refs++;
	  }
	}
      }
    }
  }
}


void
MLGraph::minimum_lcd (UINT32& min_memmem, 
		      UINT32& min_ivar, 
		      UINT32& min_scalar,
		      UINT32& memmems, 
		      UINT32& ivars, 
		      UINT32& scalars)
/////////////////////////////////////////////////////////////////////
//
// Get the minimum lcd for 1) induction variables, 2) memory 
// variables, and 3) other scalar variables.
// 
/////////////////////////////////////////////////////////////////////
{
  UINT8 opnum;
  UINT8 ixopnd;
  UINT8 omega;

  min_memmem = 32;
  min_ivar = 32;
  min_scalar = 32;
  memmems = 0;
  ivars = 0;
  scalars = 0;

  for (INT i = 0; i < m_nodes.size(); i++)
  {
    OP *op = m_nodes[i];

    for (INT j = 0; j < m_nodes.size(); j++)
    {
      MLEdge edge = m_edges[get_address(i, j)];
      ARC *arc = edge.edge;
      if (normal_edge(edge))
      {
	omega = edge.omega;
	if ((i == j) && induction_var(op, i))
	{
	  // This is an induction variable.
	  lcdmin(min_ivar, omega);
	  if (omega) ivars++;
	  continue;
	}

	if (OP_memory(op) && OP_memory(m_nodes[j]))
	{
	  // This is a memory-to-memory dependence.
	  lcdmin(min_memmem, omega);
	  if (omega) memmems++;
	  continue;
	}

	if (OP_memory(op) && 
	    (ARC_kind(arc) == CG_DEP_REGIN ||
	     ARC_kind(arc) == CG_DEP_REGANTI))
	{
	  opnum = ARC_opnd(arc);
	  ixopnd = (floating_store(op)) ? 2 : 3;
	  
	  if ((opnum == ixopnd) || induction_var(m_nodes[j], j))
	    // Don't count indices into arrays from standard induction
	    // variables as a dependence.
	    continue;
	}

	lcdmin(min_scalar, omega);
	if (omega) scalars++;
      }
    }
  }
}


BOOL
MLGraph::induction_var (OP *op, INT ix)
//////////////////////////////////////////////////////////////////
//
//  Returns true if the op is an induction variable.  ix is the
//  id of op.
//
//////////////////////////////////////////////////////////////////
{
  if (!OP_iadd(op)) return FALSE;

  if (!normal_edge_present(ix, ix))
    return FALSE;

  for (INT i = 0; i < m_nodes.size(); i++)
  {
    if (i != ix && normal_edge_present(ix, i))
      return FALSE;
  }

  return TRUE;
}


BOOL
MLGraph::floating_store (OP *op)
//////////////////////////////////////////////////////////////////
//
//  Returns true if op is a floating point store.
//
//////////////////////////////////////////////////////////////////
{
  INT opc = OP_code (op);

  switch (opc)
  {
  case TOP_stfs:
  case TOP_stfs_i:
  case TOP_stfd:
  case TOP_stfd_i:
  case TOP_stfe:
  case TOP_stfe_i:
  case TOP_stf8:
  case TOP_stf8_i:
  case TOP_stf_spill:
  case TOP_stf_i_spill:
    return TRUE;
  default:
    return FALSE;
  }
}


BOOL
MLGraph::connected_to_load (INT start)
//////////////////////////////////////////////////////////////////
//
//  Returns true if the start'th operation is connected in some
//  way to a load operation.
//
//////////////////////////////////////////////////////////////////
// Follow all paths to see if a load is encountered.
{
  for (INT i = 0; i < m_nodes.size(); i++)
  {
    if (i != start && normal_edge_present(start, i))
    {
      if (OP_load(m_nodes[i]))
	return TRUE;
      if (connected_to_load(i))
	return TRUE;
    }
  }

  return FALSE;
}


void
MLGraph::find_invariants ()
//////////////////////////////////////////////////////////////////
//
//  Finds the number of operations that are loop invariants.
//
//////////////////////////////////////////////////////////////////
{
  CG_LOOP_DEF tn_def(m_loophead);

  for (INT i = 0; i < m_nodes.size(); i++)
  {
    OP *op = m_nodes[i];
    for (INT opnd = 0; opnd < OP_opnds(op); opnd++) 
    {
      TN *tn = OP_opnd (op, opnd);
      if (tn_def.Get(tn) == NULL)
	m_invariants++;
    }
  }
}
