#include "loop_profile.h"
#include "gra_live.h"
#include "calls.h"
#include "cg_loop.h"
#include "cg.h"
#include "cgexp.h"
#include "data_layout.h"
#include "dominate.h"
#include "ipfec_options.h"
#include "region.h"
#include "region_bb_util.h"
#include "region_update.h"
#include "region_verify.h"
#include "symtab_utils.h"
#include <fstream>
#include <string>

// #define DBG_LOOP_PROFILE if you want to debug
#define ASSERT_ZERO    0
#define ASSERT_NONZERO 1

#ifdef REGISTER_PROBLEMS_
#define Set_TN_reg_value(tn, value)  (tn->u1.reg_tn.class_reg.class_reg.reg = value)
#endif

char *LOOP_INSTRUMENT::LOOP_TIMERS_ARR   = "_loop_timers";
char *LOOP_INSTRUMENT::LOOP_TIMERS_SIZE  = "_loop_timers_size";
char *LOOP_INSTRUMENT::LOOP_TIMERS_ERROR = "_loop_timers_error";

void 
CG_Loop_Instrument (CGRIN* rin, PROFILE_PHASE phase, BOOL add_checks)
/////////////////////////////////////////////////////////////////
// 
//  The interface to the library.
//
/////////////////////////////////////////////////////////////////
{
  LOOP_INSTRUMENT li(rin, phase, add_checks);

#ifdef DBG_LOOP_PROFILE
  fprintf (stderr, "CG_Loop_Instrument: %s %s %s.\n", 
	   Src_File_Name, Cur_PU_Name,
	   add_checks ? "CORRECTNESS MODE" : "FAST MODE");
#endif

  li.Instrument ();

#ifdef DBG_LOOP_PROFILE
  fprintf (stderr, "Exiting CG_Loop_Instrument pass.\n");
#define
}


LOOP_INSTRUMENT::LOOP_INSTRUMENT (CGRIN *rin, 
				  PROFILE_PHASE phase,
				  BOOL checks) :
  m_rin(rin),
  m_phase(phase),
  m_measure_outermost(FALSE),
  m_perform_checks (checks)
/////////////////////////////////////////////////////////////////
// 
//  The constructor.
// 
/////////////////////////////////////////////////////////////////
{
  MEM_POOL_Initialize (&m_mempool, "LOOP_PROFILE_MEM", true);
  MEM_POOL_Push (&m_mempool);
  m_srcfile_pu_name = CXX_NEW_ARRAY(char, 
				    strlen(Src_File_Name) + 
				    strlen(Cur_PU_Name) + 2,
				    &m_mempool);
  strcpy (m_srcfile_pu_name, Src_File_Name);
  strcat (m_srcfile_pu_name, "/");
  strcat (m_srcfile_pu_name, Cur_PU_Name);

  Create_Global_Symbols ();
}

LOOP_INSTRUMENT::~LOOP_INSTRUMENT ()
/////////////////////////////////////////////////////////////////
// 
//  The destructor.
// 
/////////////////////////////////////////////////////////////////
{
  MEM_POOL_Pop (&m_mempool);
  MEM_POOL_Delete (&m_mempool);
}

bool
LOOP_INSTRUMENT::Skip_PU ()
/////////////////////////////////////////////////////////////////
// 
//  This function is for debugging purposes.  If the file
//  skipfile.txt exists in the directory that the compiler is
//  invoked from, procedure names matching the names in the file
//  will not be processed.
// 
/////////////////////////////////////////////////////////////////
{
  ifstream skipfile("skipfile.txt");
  string procname;

  while (skipfile >> procname)
  {
    if (!strcmp(procname.c_str(), Cur_PU_Name))
      return true;
  }
  return false;
}

void
LOOP_INSTRUMENT::Instrument ()
/////////////////////////////////////////////////////////////////
//
//  This function does most of the work.  It identifies loops,  
//  then creates a prolog and epilogs for each of the inner-
//  most loops.  Code is inserted in the prolog that starts
//  a timer, and likewise, code is inserted in the epilogues
//  that stop the timer.
//
/////////////////////////////////////////////////////////////////
{
  LOOP_DESCR *outer;
  LOOP_DESCR *outermost;
  INT32 scid;
  INT gid;
  BB *loophead;

  if (Skip_PU()) 
  {
    // Skip any procedure whose name is in the file skipfile.txt
    fprintf (stderr, "WARNING: Skipping %s...\n", Cur_PU_Name);
    return;
  }

  CXX_MEM_POOL local_mp("LICM", FALSE);
  
  MEM_POOL loop_descr_pool;
  MEM_POOL_Initialize (&loop_descr_pool, "loop_descriptors", TRUE);
  MEM_POOL_Push (&loop_descr_pool);
  
#ifdef DBG_LOOP_PROFILE
  Dot_Print ("pre");
#endif

  // Make sure liveness information is accurate.
  GRA_LIVE_Init (NULL);
  Calculate_Dominators ();

  // Tag each loop with a unique ID.  The IDs are assigned in
  // increasing order, starting with zero.  The last ID to 
  // be used is stored in last_instr_gid
  gid = read_gid ();
  for (LOOP_DESCR *loop = LOOP_DESCR_Detect_Loops(&loop_descr_pool);
       loop != NULL;
       loop = LOOP_DESCR_next(loop)) 
  {
    if (BB_innermost(LOOP_DESCR_loophead(loop)) &&
	!LOOP_DESCR_Has_Side_Entrance(loop) &&
	!Skip_Loop_For_Reason(loop))
    {
      // Write some other info about the loop out to disk so
      // we can later associate the gid with the actual loop.
      scid = BB_Loop_Lineno(LOOP_DESCR_loophead(loop));

      loophead = LOOP_DESCR_loophead(loop);
      write_associations (m_srcfile_pu_name, scid, gid, BB_id(loophead));

      // Annotate the loophead with the id of this loop.
      BB_Add_Annotation(loophead, ANNOT_LOOPID, (void*)gid);

      if ((CG_Loop_Optimization != 50) && 
	  (CG_Loop_Optimization != 51))
      {
	// Skip the instrumentation part if we're going to 
	// back-annotate.
	Create_Instrumented_Prolog (loop, gid);
	Create_Instrumented_Epilog (loop, gid);
      }
      gid++;
    }
  }

#ifdef DBG_LOOP_PROFILE
  Dot_Print ("post");
#endif

  // Store the gid back out to disk.
  write_gid (gid);

  GRA_LIVE_Init (NULL);
  Free_Dominators_Memory ();
  
  MEM_POOL_Pop (&loop_descr_pool);
  MEM_POOL_Delete(&loop_descr_pool);
}


void
LOOP_INSTRUMENT::Create_Instrumented_Prolog (LOOP_DESCR *loop, INT id)
/////////////////////////////////////////////////////////////////
//
//  Create a prolog for the loop, then insert the instru-
//  mentation code.
//
/////////////////////////////////////////////////////////////////
{
  BB *prolog;
  BB *unique_succ;

  prolog = Create_Prolog (loop);
#ifdef DBG_LOOP_PROFILE
  fprintf (stderr, "LOOP_INSTRUMENT adding prolog BB%d\n", BB_id(prolog));
#endif
  
  // If the prolog has branches, then even though it may not
  // be necessary -- there are all sorts of assumptions that
  // downstream passes make -- I feel much better having a 
  // dedicated block for the instrumented prolog.
  if (BB_xfer_op(prolog))
  {
    fprintf (stderr, "ATTENTION: branches at prolog's end.\n");
    prolog = Create_Empty_Prolog (loop, prolog);
  }
  
  unique_succ = BB_Unique_Successor (prolog);
  assert (unique_succ);

  Start_Timer (prolog, id);
}


BB*
LOOP_INSTRUMENT::Create_Prolog (LOOP_DESCR *l)
/////////////////////////////////////////////////////////////////
//
//  Creates a prolog for the loop that is passed in.  If there
//  is already a block that precedes the loop that resembles
//  a prolog, return that; otherwise, create a basic block
//  and prepend it to the head of the loop.
// 
/////////////////////////////////////////////////////////////////
{
  BB* lhead = LOOP_DESCR_loophead(l);
  BB_SET* body = LOOP_DESCR_bbset(l);
  BB* like_prolog = NULL;

  BBLIST* p;
  FOR_ALL_BB_PREDS (lhead, p) 
  {
    BB* pred = BBLIST_item (p);
    if (BB_SET_MemberP (body, pred))
      continue;
    
    if (like_prolog) 
    {
      like_prolog = NULL; 
      break; 
    } 
    else
      like_prolog = pred;
  }

  // Generate a new prolog if any of the following conditions
  // are true.
  if (like_prolog &&
      (BB_kind(like_prolog) != BBKIND_GOTO || 
       BB_branch_op(like_prolog) ||
       BB_entry(like_prolog) ||
       BB_call(like_prolog) ||
       BB_xfer_op(like_prolog)))
    like_prolog = NULL;
  
  if (like_prolog && !CG_localize_tns) 
    Set_BB_gra_spill(like_prolog);

  Is_True (!like_prolog || BB_prev(lhead) == like_prolog, 
	   ("BB:%d's does not fall through to BB:%d", 
	    BB_id(like_prolog), BB_id(lhead)));
  
  return like_prolog ? 
    like_prolog : CG_LOOP_Gen_And_Prepend_To_Prolog (lhead, l);
}


BB*
LOOP_INSTRUMENT::Create_Empty_Prolog (LOOP_DESCR *loop, BB *old_prolog)
/////////////////////////////////////////////////////////////////
//
//  Create an empty prolog block and patch it up in between
//  the <old_prolog> with a branch at the end, and the loop
//  head.
//
/////////////////////////////////////////////////////////////////
{
  return old_prolog;
}


void
LOOP_INSTRUMENT::Create_Instrumented_Epilog (LOOP_DESCR *loop, INT id)
/////////////////////////////////////////////////////////////////
//
//  Create an epilog for the loop, then insert the instru-
//  mentation code.
//
/////////////////////////////////////////////////////////////////
{
  BB_ASSOCS epilog_map;  
  BB_ASSOCS::iterator first;
  BB *epilog;

  // Creates the epilogs and stores the epilogs, as well
  // as the original exit targets in the <epilog_map>.
  Create_Epilogs (loop, epilog_map);

  for (first = epilog_map.begin(); 
       first != epilog_map.end();
       first++)
  {
    epilog = (*first).second;
    Stop_Timer (epilog, id);
  }
}


void
LOOP_INSTRUMENT::Create_Epilogs (LOOP_DESCR *l, BB_ASSOCS& epilog_map)
/////////////////////////////////////////////////////////////////
//
//  Creates an epilog block for each of the exits from the
//  loop.  If all exits target the same block, only one
//  epilog block is created.  The map <epilog_map> that is 
//  passed in is used to associate each newly created epi-
//  log with the block that the loop originally targetted.
//  We use the map to return the epilog blocks that this
//  procedure creates.
// 
/////////////////////////////////////////////////////////////////
{
  BB_ASSOCS non_loop_fts;
  EDGE_LIST exit_edges;

  // Create a new block for all of the targets.  If two exits
  // share the same target, only create one new block.
  Create_Epilog_Blocks (epilog_map, non_loop_fts, exit_edges, l);

  // Insert the newly created epilogs into the graph.
  Insert_Epilog_Blocks (epilog_map, non_loop_fts, exit_edges, l);
}

void
LOOP_INSTRUMENT::Create_Epilog_Blocks (BB_ASSOCS& epilog_map,
				       BB_ASSOCS& non_loop_fts,
				       EDGE_LIST& exit_edges,
				       LOOP_DESCR *l)
/////////////////////////////////////////////////////////////////
//
//  Returns the basic block that all exits in the loop target.
//  If there isn't a common target block, NULL is returned.
//  Here we also do a lot of bookkeeping; we keep track of the
//  exit edges from the loop to their non-loop targets, the
//  targets that are actually fall-through targets for non-
//  loop basic blocks, and the new epilog blocks that we 
//  create for each loop target.
// 
/////////////////////////////////////////////////////////////////
{
  BBLIST *succs;
  BB *loop_node;
  BB *target;
  BB *nlft;
  EDGE_INFO tmp;

  FOR_ALL_BB_SET_members (LOOP_DESCR_bbset(l), loop_node)
  {
    FOR_ALL_BB_SUCCS (loop_node, succs)
    {
      target = BBLIST_item (succs);
      if (!BB_SET_MemberP(LOOP_DESCR_bbset(l), target))
      {
	// We found a successor to a loop_node that's not
	// a member of the loop.  Thus, it's a loop exit.
	if (epilog_map.find(target) == epilog_map.end())
	{
	  // This target doesn't have an associated
	  // epilog block yet, so let's make one. We'll
	  // actually insert it into the graph later.
	  epilog_map[target] = Gen_And_Set_BB_Like (target);
#ifdef DBG_LOOP_PROFILE
	  fprintf (stderr, "LOOP_INSTRUMENT adding epilog BB%d\n", 
		   BB_id(epilog_map[target]));
#endif

	  // If there are non-loop, fall-thru predecessors
	  // of <target>, remember the non-loop node.
	  nlft = Non_Loop_Fall_Thru_Pred (target, l);
	  if (nlft)
	    non_loop_fts[target] = nlft;
	}

	// Save the exit edge information away.  We want
	// to find out now whether the edge was originally
	// a fall-through edge, as the information will
	// change when we start manipulating the graph.
	tmp.exit_node = loop_node;
	tmp.target_node = target;
	tmp.fall_through = BB_Fall_Thru_Successor (loop_node);
	exit_edges.push_back (tmp);
      }
    }
  }
}


void
LOOP_INSTRUMENT::Insert_Epilog_Blocks (BB_ASSOCS& epilog_map,
				       BB_ASSOCS& non_loop_fts,
				       EDGE_LIST& exit_edges,
				       LOOP_DESCR *l)
/////////////////////////////////////////////////////////////////
//
//  This function finds all the exits from the loop and 
//  inserts a new basic block between all loop exits and the
//  blocks that they target.  We use the bookkeeping that was
//  created in Create_Epilog_Blocks.
// 
/////////////////////////////////////////////////////////////////
{
  BB *nlft;
  BB *target;
  BB *prolog;
  BB *brfixup;
  EDGE_INFO edge;
  BB_ASSOCS::iterator ft_iter;
  BB_ASSOCS::iterator first_epilog = epilog_map.begin ();
  EDGE_LIST::iterator first_edge = exit_edges.begin ();

  while (first_epilog != epilog_map.end())
  {
    target = (*first_epilog).first;
    prolog = (*first_epilog).second;

    // Insert the prolog before the target.
    BB_next (prolog) = target;
    BB_prev (prolog) = BB_prev (target);
    if (BB_prev(target))
      BB_next(BB_prev(target)) = prolog;
    BB_prev(target) = prolog;
    Link_Pred_Succ_with_Prob (prolog, target, 1.0);

    // If there's a non-loop node that falls through to <target>,
    // we have to insert a goto from that node to <target> (since
    // there is now a prolog block that falls through to <target>).
    ft_iter = non_loop_fts.find (target);
    if (ft_iter != non_loop_fts.end())
    {
      nlft = (*ft_iter).second;
      fprintf (stderr, "ATTENTION: found a non-loop fall through (BB%d).\n",
	       BB_id(nlft));
      
      if (BB_branch_op(nlft)) 
      {
	// There is a branch in <nlft> but <target> was the
	// fall through node.  Since <prolog> is now 
	// in front of <target> we have to branch around
	// <prolog>.  Therefore, we insert yet another
	// block after <nlft> (and before <prolog>) and
	// insert a goto from <brfixup> to <target>.
	brfixup = Gen_And_Insert_BB_After (nlft);
	Change_Succ (nlft, target, brfixup);
	Add_Goto (brfixup, target);
      }
      else
      {
	// This is a true fall through node, so insert a goto
	// in <nlft>.
	Add_Goto (nlft, target);
      }
    }
    
    first_epilog++;
  }
  
  // Fixup the branches and the pred-succ links. 
  while (first_edge != exit_edges.end())
  {
    edge = *first_edge;
    if (edge.fall_through == edge.target_node)
    {
      // The <target> node was the fall-through node of
      // <exit> so we just need to fix the pred-succ links.
      Unlink_Pred_Succ (edge.exit_node, edge.target_node);
      Link_Pred_Succ (edge.exit_node, epilog_map[edge.target_node]);
    }
    else
    {
      // The <target> node was not the fall through node
      // so we have to retarget the branch in
      // <exit_node> to go to <prolog> instead of <target>. 
      assert (BB_Retarget_Branch(edge.exit_node, edge.target_node,
				 epilog_map[edge.target_node]));
    }

    first_edge++;
  }
}


BB*
LOOP_INSTRUMENT::Non_Loop_Fall_Thru_Pred (BB *target,
					  LOOP_DESCR *l)
/////////////////////////////////////////////////////////////////
//
//  If the loop target is preceeded in code-space by a block
//  that is not part of the loop, return that block.  If this
//  is the case, then we have to fixup the CFG so that the 
//  non-loop fall through jumps to loop target instead of the
//  epilog that we are going to insert in front of the
//  target.
//
/////////////////////////////////////////////////////////////////
{
  BBLIST *preds;
  BB *node;

  FOR_ALL_BB_PREDS (target, preds)
  { 
    node = BBLIST_item (preds);
    if (!BB_SET_MemberP(LOOP_DESCR_bbset(l), node))
      if (BB_Fall_Thru_Successor(node) == target)
	return node;
  }

  return NULL;
}

 
ST*
LOOP_INSTRUMENT::Get_Global_ST (char* symname)
/////////////////////////////////////////////////////////////////
//
//  Return the global symtab that matches symname. If one does
//  not exist, make it.
//
/////////////////////////////////////////////////////////////////
{
  INT i;
  ST *global_st;
  TY_IDX ty;

  FOREACH_SYMBOL (GLOBAL_SYMTAB, global_st, i) {
    if ((*ST_name(global_st) == *symname) && 
	(strcmp(ST_name(global_st), symname) == 0))
      return global_st;
  }

  // A global symtab was not found, so let's create one.
  ty = MTYPE_To_TY (MTYPE_I8);
  global_st = New_ST (GLOBAL_SYMTAB);
    
  ST_Init (global_st, Save_Str(symname), CLASS_VAR, 
	   SCLASS_EXTERN, EXPORT_PREEMPTIBLE, ty);
  Allocate_Object (global_st);  

  return global_st;
}


void
LOOP_INSTRUMENT::Create_Global_Symbols ()
/////////////////////////////////////////////////////////////////
//
//  This function creates some global symbols that we'll
//  access throughout the class.
//
/////////////////////////////////////////////////////////////////
{
  TY_IDX ty;
  ST *loop_timers_st;
  ST *timers_size_st;
  ST *error_st;

  loop_timers_st = Get_Global_ST (LOOP_TIMERS_ARR);
  assert (loop_timers_st);
  m_time_arr_base_offset = Gen_Symbol_TN (loop_timers_st, 
					  0, TN_RELOC_IA_GPREL22);

  timers_size_st = Get_Global_ST (LOOP_TIMERS_SIZE);
  assert (timers_size_st);
  m_time_size_offset = Gen_Symbol_TN (timers_size_st,
				      0, TN_RELOC_IA_GPREL22);

  error_st = Get_Global_ST (LOOP_TIMERS_ERROR);
  assert (error_st);
  m_error_offset = Gen_Symbol_TN (error_st,
				  0, TN_RELOC_IA_GPREL22);

  // Build a TN to describe the ar.itc register.
  m_itc_tn = Build_Dedicated_TN (ISA_REGISTER_CLASS_application,
				 45, 8);
#ifdef REGISTER_PROBLEMS_
  m_stack_reg = REGISTER_Request_Stacked_Register (ABI_PROPERTY_caller, 
						   ISA_REGISTER_CLASS_integer);
#endif
}


void
LOOP_INSTRUMENT::Start_Timer (BB *bb, INT id)
/////////////////////////////////////////////////////////////////
//
//  Instruments the basic block that is passed in by creating
//  the code necessary to capture the current 'ticks' and
//  store them away in a global variable (to the compiled
//  program).
//
/////////////////////////////////////////////////////////////////
{
  TN *arr_ptr;
  TN *ticks;
  TN *stored_ticks;
  TN *stored_ticks_addr;
  OP *addl_op;
  OP *get_ticks;
  OP *ld_op;
  OP *id_op;
  OP *st_op;
  OP *diff_op;
  OP *add_op;
  OP *shladd_op;

  // Get the base offset: arr_ptr=@gprel(_loop_timers#),gp
  arr_ptr = Gen_Register_TN (ISA_REGISTER_CLASS_integer, Pointer_Size);
  addl_op = Mk_OP (TOP_addl, arr_ptr, True_TN, m_time_arr_base_offset, GP_TN);
  Set_OP_no_move_before_gra (addl_op);
  BB_Append_Op (bb, addl_op);

  // Now, point to the base of the address stored at arr_ptr.
  ld_op = Mk_OP (TOP_ld8, arr_ptr, True_TN,
		 Gen_Enum_TN(ECV_ldtype),
		 Gen_Enum_TN(ECV_ldhint), arr_ptr);
  //  Set_OP_volatile (ld_op);
  Set_OP_no_move_before_gra (ld_op);
  BB_Append_Op (bb, ld_op);

  // Compute the index into the array: stored_ticks_addr=id << 4 + arr_ptr
  stored_ticks_addr = Gen_Register_TN (ISA_REGISTER_CLASS_integer, Pointer_Size);
  id_op = Mk_OP (TOP_mov_i, stored_ticks_addr, True_TN, Gen_Literal_TN(id, 4));
  Set_OP_no_move_before_gra (id_op);
  BB_Append_Op (bb, id_op);

  shladd_op = Mk_OP (TOP_shladd, stored_ticks_addr, True_TN, stored_ticks_addr, 
 		     Gen_Literal_TN(4, 4), arr_ptr);
  Set_OP_no_move_before_gra (shladd_op);
  BB_Append_Op (bb, shladd_op);

  // Inserts instructions to make sure that the value in the 
  // ticks array is zero.  This is an invariant for the
  // start timer because Stop_Timer sets it equal to zero
  // as soon as it records the elapsed time and adds it in
  // to the cumulative array.
  Consistency_Check (bb, stored_ticks_addr, ASSERT_ZERO, id);

  // Get the ar.itc value in a register: ticks=ar.itc
  ticks = Gen_Register_TN (ISA_REGISTER_CLASS_integer, 8);
  get_ticks = Mk_OP (TOP_mov_f_ar, ticks, True_TN, m_itc_tn);
  Set_OP_no_move_before_gra (get_ticks);
  //  Set_OP_volatile (get_ticks);
  BB_Append_Op (bb, get_ticks);

  // Move the ticks counter value into the ticks array.
  st_op = Mk_OP (TOP_st8, True_TN,
		 Gen_Enum_TN(ECV_sttype),
		 Gen_Enum_TN(ECV_sthint), stored_ticks_addr,
		 ticks);
  //  Set_OP_volatile (st_op);
  Set_OP_no_move_before_gra (st_op);
  BB_Append_Op (bb, st_op);  

#ifdef DBG_LOOP_PROFILE
  Print_BB (bb);
#endif 

  // Recalculate liveness information.
  GRA_LIVE_Recalc_Liveness (NULL);
}

void
LOOP_INSTRUMENT::Stop_Timer (BB *bb, INT id)
/////////////////////////////////////////////////////////////////
//
//  Instruments the basic block that is passed in by creating
//  the code necessary to capture the current 'ticks' and
//  see how many ticks have elapsed since the timer was 
//  started.
//
/////////////////////////////////////////////////////////////////
{
  TN *arr_ptr;
  TN *ticks;
  TN *stored_ticks;
  TN *cum_ticks;
  TN *cum_ticks_addr;
  TN *stored_ticks_addr;
  OP *addl_op;
  OP *get_ticks;
  OP *ld_op;
  OP *id_op;
  OP *st_op;
  OP *diff_op;
  OP *add_op;
  OP *shladd_op;

  // Get the ar.itc value in a register: ticks=ar.itc
  ticks = Gen_Register_TN (ISA_REGISTER_CLASS_integer, 8);
  get_ticks = Mk_OP (TOP_mov_f_ar, ticks, True_TN, m_itc_tn);
  Set_OP_no_move_before_gra (get_ticks);
  //  Set_OP_volatile (get_ticks);
  BB_Append_Op (bb, get_ticks);

  // Get the base offset: arr_ptr=@gprel(_loop_timers#),gp
  arr_ptr = Gen_Register_TN (ISA_REGISTER_CLASS_integer, Pointer_Size);
  addl_op = Mk_OP (TOP_addl, arr_ptr, True_TN, m_time_arr_base_offset, GP_TN);
  Set_OP_no_move_before_gra (addl_op);
  BB_Append_Op (bb, addl_op);

  // Now, point to the base of the address stored at arr_ptr.
  ld_op = Mk_OP (TOP_ld8, arr_ptr, True_TN,
		 Gen_Enum_TN(ECV_ldtype),
		 Gen_Enum_TN(ECV_ldhint), arr_ptr);
  //  Set_OP_volatile (ld_op);
  Set_OP_no_move_before_gra (ld_op);
  BB_Append_Op (bb, ld_op);

  // Compute the index into the array: stored_ticks_addr=id << 4 + arr_ptr
  stored_ticks_addr = Gen_Register_TN (ISA_REGISTER_CLASS_integer, Pointer_Size);
  id_op = Mk_OP (TOP_mov_i, stored_ticks_addr, True_TN, Gen_Literal_TN(id, 4));
  Set_OP_no_move_before_gra (id_op);
  BB_Append_Op (bb, id_op);

  shladd_op = Mk_OP (TOP_shladd, stored_ticks_addr, True_TN, stored_ticks_addr, 
 		     Gen_Literal_TN(4, 4), arr_ptr);
  Set_OP_no_move_before_gra (shladd_op);
  BB_Append_Op (bb, shladd_op);

  // Inserts instructions to make sure that the value in the 
  // ticks array is zero.  This is an invariant for the
  // start timer because Stop_Timer sets it equal to zero
  // as soon as it records the elapsed time and adds it in
  // to the cumulative array.
  Consistency_Check (bb, stored_ticks_addr, ASSERT_NONZERO, id);
    
  // Load the ticks from the array: stored_ticks=[stored_ticks_addr]
  stored_ticks = Gen_Register_TN (ISA_REGISTER_CLASS_integer, 8);
  ld_op = Mk_OP (TOP_ld8, stored_ticks, True_TN, 
 		 Gen_Enum_TN(ECV_ldtype),
 		 Gen_Enum_TN(ECV_ldhint), stored_ticks_addr);
  //  Set_OP_volatile (ld_op);
  Set_OP_no_move_before_gra (ld_op);
  BB_Append_Op (bb, ld_op);

  // Get the offset to the cum_ticks field: cum_ticks_addr=stored_ticks_addr+8
  cum_ticks_addr = Gen_Register_TN (ISA_REGISTER_CLASS_integer, Pointer_Size);
  add_op = Mk_OP (TOP_add, cum_ticks_addr, True_TN, stored_ticks_addr, 
		  Gen_Literal_TN(8, 4));
  Set_OP_no_move_before_gra (add_op);
  BB_Append_Op (bb, add_op);

  // Get the cumulative running time of the loop: cum_ticks=[cum_ticks_addr]
  cum_ticks = Gen_Register_TN (ISA_REGISTER_CLASS_integer, 8);
  ld_op = Mk_OP (TOP_ld8, cum_ticks, True_TN,
		 Gen_Enum_TN(ECV_ldtype),
		 Gen_Enum_TN(ECV_ldhint), cum_ticks_addr);
  //  Set_OP_volatile (ld_op);
  Set_OP_no_move_before_gra (ld_op);
  BB_Append_Op (bb, ld_op);

  // Find out how long the loop took to run: ticks=ticks-stored_ticks
  diff_op = Mk_OP (TOP_sub, ticks, True_TN, ticks, stored_ticks);
  Set_OP_no_move_before_gra (diff_op);
  BB_Append_Op (bb, diff_op);

  // Add this value to the cumulative running time: cum_ticks += ticks
  add_op = Mk_OP (TOP_add, cum_ticks, True_TN, cum_ticks, ticks);
  Set_OP_no_move_before_gra (add_op);
  BB_Append_Op (bb, add_op);

  // Store this value away: [cum_ticks_addr]=cum_ticks
  st_op = Mk_OP (TOP_st8, True_TN,
		 Gen_Enum_TN(ECV_sttype),
		 Gen_Enum_TN(ECV_sthint), cum_ticks_addr,
		 cum_ticks);
  //  Set_OP_volatile (st_op);
  Set_OP_no_move_before_gra (st_op);
  BB_Append_Op (bb, st_op);  

  // Put a 0 in the ticks entry: [stored_ticks_addr]=0
  st_op = Mk_OP (TOP_st8, True_TN,
		 Gen_Enum_TN(ECV_sttype),
		 Gen_Enum_TN(ECV_sthint), stored_ticks_addr,
		 Zero_TN);
  //  Set_OP_volatile (st_op);
  Set_OP_no_move_before_gra (st_op);
  BB_Append_Op (bb, st_op);  

#ifdef DBG_LOOP_PROFILE
  Print_BB (bb);
#endif

  // Recalculate liveness information.
  GRA_LIVE_Recalc_Liveness (NULL);
}


void
LOOP_INSTRUMENT::Consistency_Check (BB *bb, 
				    TN *ticks_addr, 
				    INT check,
				    INT id)
/////////////////////////////////////////////////////////////////
//
//  Checks to make sure that the data at <ticks_addr> is 
//  either zero or non-zero, depending on the state of <check>.
//  The idea here is this: Stop_Timer places a zero in the
//  ticks array (the index of which is determined by the loop
//  that it is working on), so when Start_Timer is called for
//  a particular loop, its ticks value in the ticks array had
//  better be zero.  Likewise, Start_Timer places the current
//  itc value in the ticks array, so when Stop_Timer is called
//  it had better not be zero.
//
//  Note that there is a very, very small chance that the value
//  of ar.itc will be zero when it is read by Start_Timer.  We
//  do not check for this condition.
//
/////////////////////////////////////////////////////////////////
{
  if (m_perform_checks)
  {
    TN *ticks_val;
    TN *err_ptr;
    TN *id_tn;
    TN *pred_tn;
    OP *addl_op;
    OP *ld_op;
    OP *id_op;
    OP *cmp_op;
    OP *st_op;
    TOP comparison;

    // Load the ticks value for the loop we're working on.
    ticks_val = Gen_Register_TN (ISA_REGISTER_CLASS_integer, 8);
    ld_op = Mk_OP (TOP_ld8, ticks_val, True_TN, 
		   Gen_Enum_TN(ECV_ldtype),
		   Gen_Enum_TN(ECV_ldhint), ticks_addr);
    //    Set_OP_volatile (ld_op);
    Set_OP_no_move_before_gra (ld_op);    
    BB_Append_Op (bb, ld_op);

    // Figure out what kind of comparison we want to perform.
    comparison = (check == ASSERT_ZERO) ? TOP_cmp_eq : TOP_cmp_ne;

    // Perform a comparison: compare p0,pred=ticks_val,r0
    pred_tn = Build_RCLASS_TN (ISA_REGISTER_CLASS_predicate);
    cmp_op = Mk_OP (comparison, True_TN, pred_tn, True_TN, 
		    ticks_val, Zero_TN);
    Set_OP_no_move_before_gra (cmp_op);    
    BB_Append_Op (bb, cmp_op);
    
    // Find out where the _loop_timers_error variable lives.
    err_ptr = Gen_Register_TN (ISA_REGISTER_CLASS_integer, 
			       Pointer_Size);
    addl_op = Mk_OP (TOP_addl, err_ptr, True_TN, 
		     m_error_offset, GP_TN);
    Set_OP_no_move_before_gra (addl_op);    
    BB_Append_Op (bb, addl_op);

    // Get the ID into a register.
    id_tn = Gen_Register_TN (ISA_REGISTER_CLASS_integer, 8);
    id_op = Mk_OP (TOP_mov_i, id_tn, pred_tn, Gen_Literal_TN(id, 4));
    Set_OP_no_move_before_gra (id_op);    
    BB_Append_Op (bb, id_op);

    // Store the id in the _loop_timers_error variable,
    // predicated on pX.
    st_op = Mk_OP (TOP_st8, pred_tn,
		   Gen_Enum_TN(ECV_sttype),
		   Gen_Enum_TN(ECV_sthint), err_ptr,
		   id_tn);
    //    Set_OP_volatile (st_op);
    Set_OP_no_move_before_gra (st_op);
    BB_Append_Op (bb, st_op);  
  }
}


BB*
LOOP_INSTRUMENT::Gen_And_Set_BB_Like (BB *model)
/////////////////////////////////////////////////////////////////
//
//  Generates a new BB like <model>.  In other words, it 
//  creates a BB with the same RID as model.  It also 
//  sets the GRA spill flag.  What the purpose of this is,
//  I really can't say; ORC is undocumented.  The only reason
//  I'm setting this flag is because other passes seem to do
//  it when new blocks are added.
//
/////////////////////////////////////////////////////////////////
{
  BB* bb;

  bb = Gen_BB_Like (model);
  if (!CG_localize_tns) 
    Set_BB_gra_spill (bb);

  return bb;
}


void
LOOP_INSTRUMENT::Print_BB (BB *bb)
/////////////////////////////////////////////////////////////////
//  For debugging.  Prints the basic block that is passed in.
/////////////////////////////////////////////////////////////////
{
  BBLIST *bl;
  
  printf ("****** Begin Block %d: Preds {", bb->id);
  FOR_ALL_BB_PREDS (bb, bl) {
    printf (" %d", BB_id(BBLIST_item(bl)));
  }
  printf ("} Succs {" );
  FOR_ALL_BB_SUCCS (bb, bl) {
    printf (" %d", BB_id(BBLIST_item(bl)));
  }
  printf ("}\n");
  
  op *op = BB_first_op (bb);
  if (op)
  {
    while (op != BB_last_op(bb))
    {
      Print_OP_No_SrcLine (op);
      op = OP_next (op);
    }
    Print_OP_No_SrcLine (op);
  }

  printf ("****** End Block %d:\n", bb->id);
}


void
Dot_Print (char *spec)
/////////////////////////////////////////////////////////////////
//  For debugging.  Prints the CFG in dot format.
/////////////////////////////////////////////////////////////////
{
  ANNOTATION *annot;
  BB_LIST *bblist;
  INT loopid;
  BB *entry;

  BBLIST *bl;
  char dotfname[80];

  sprintf (dotfname, "%s_%s.dot", spec, Cur_PU_Name);
  FILE *dotfile = fopen (dotfname, "w");
  fprintf (dotfile, "digraph %s_%s {\n", spec, Cur_PU_Name);

  for (bblist = Entry_BB_Head; bblist; bblist = BB_LIST_rest(bblist))
  {
    entry = BB_LIST_first (bblist);
    for (BB *cur = entry; cur != NULL; cur = cur->next)
    {
      annot = ANNOT_Get (BB_annotations(cur), ANNOT_LOOPINSTR);
      if (annot) {
	loopid = (INT) ANNOT_info (annot);
	fprintf (dotfile, "B%d [label=\"B%d %d\" color=blue]\n",
		 BB_id(cur), BB_id(cur), loopid);
      }

      //Print_BB(cur);
      FOR_ALL_BB_SUCCS (cur, bl) {
	fprintf (dotfile, "B%d -> B%d\n", BB_id(cur), BB_id(BBLIST_item(bl)));
      }
    }
  }

  fprintf (dotfile, "}\n");
  fclose (dotfile);
}


int
LOOP_INSTRUMENT::read_gid ()
/////////////////////////////////////////////////////////////////
//  Reads the last global loop id.
/////////////////////////////////////////////////////////////////
{
  INT id;
  FILE *gid_fname;

  gid_fname = fopen ("last_instr_gid", "r");
  if (gid_fname)
  {
    fscanf (gid_fname, "%d", &id);
    fclose (gid_fname);
  }
  else
  {
    id = 0;
  }

  return id;
}


void
LOOP_INSTRUMENT::write_gid (INT id)
/////////////////////////////////////////////////////////////////
//  Writes the GID to file.
/////////////////////////////////////////////////////////////////
{
  FILE *gid_fname;

  gid_fname = fopen ("last_instr_gid", "w");
  assert (gid_fname);
  fprintf (gid_fname, "%d\n", id);
  fclose (gid_fname);
}



void
LOOP_INSTRUMENT::write_associations (char *src_pu_name,
				     INT lid, INT gid, INT lhid)
/////////////////////////////////////////////////////////////////
//  Writes an index of loops to disk for later reference.
/////////////////////////////////////////////////////////////////
{
  FILE *assoc_fname;

  assoc_fname = fopen ("loop_instr_assocs", "a");
  assert (assoc_fname);
  fprintf (assoc_fname, "GID(%d) LOC(%s,%d) LHID(%d)\n", 
	   gid, src_pu_name, lid, lhid);
  fclose (assoc_fname);
}
